@@ -25,9 +25,11 @@ import distro_info
 import glob
 import logging
 import os
+import pwd
 import re
 import hashlib
 import subprocess
+import pathlib
 from subprocess import PIPE, Popen
 
 from .utils import get_arch
@@ -55,6 +57,7 @@ class DistUpgradeQuirks(object):
         self._snap_list = None
         self._from_version = None
         self._to_version = None
+        self._did_change_font = False
 
     # the quirk function have the name:
     #  $Name (e.g. PostUpgrade)
@@ -111,18 +114,21 @@ class DistUpgradeQuirks(object):
         logging.debug("running Quirks.PreCacheOpen")
 
     # individual quirks handler that run *after* the cache is opened
-    def kineticPostInitialUpdate(self):
+    def manticPostInitialUpdate(self):
         # PreCacheOpen would be better but controller.abort fails terribly
         """ run after the apt cache is opened the first time """
-        logging.debug("running Quirks.kineticPostInitialUpdate")
+        logging.debug("running Quirks.manticPostInitialUpdate")
         self._get_from_and_to_version()
         self._test_and_fail_on_i386()
         self._test_and_fail_on_aufs()
         self._test_and_fail_on_power8()
+        self._test_and_fail_on_bios_with_xfs_boot()
 
         cache = self.controller.cache
         self._test_and_warn_if_ros_installed(cache)
 
+        self._maybe_prevent_flatpak_auto_removal()
+
         if 'snapd' not in cache:
             logging.debug("package required for Quirk not in cache")
             return
@@ -136,8 +142,8 @@ class DistUpgradeQuirks(object):
         if self._snapstore_reachable:
             self._calculateSnapSizeRequirements()
 
-    def kineticPostUpgrade(self):
-        logging.debug("running Quirks.kineticPostUpgrade")
+    def manticPostUpgrade(self):
+        logging.debug("running Quirks.manticPostUpgrade")
         cache = self.controller.cache
         if 'snapd' not in cache:
             logging.debug("package required for Quirk not in cache")
@@ -166,6 +172,7 @@ class DistUpgradeQuirks(object):
         self._killKBluetooth()
         self._pokeScreensaver()
         self._stopDocvertConverter()
+        self._set_generic_font()
 
     # individual quirks handler that run *right before* the dist-upgrade
     # is calculated in the cache
@@ -469,6 +476,74 @@ class DistUpgradeQuirks(object):
             if not res:
                 self.controller.abort()
 
+    def _maybe_prevent_flatpak_auto_removal(self):
+        """
+        If flatpak is installed, and there are either active remotes, or
+        flatpak apps installed, prevent flatpak's auto-removal on upgrade.
+        """
+        prevent_auto_removal = False
+
+        if "flatpak" not in self.controller.cache:
+            return
+
+        if not self.controller.cache["flatpak"].is_installed:
+            return
+
+        if not os.path.exists("/usr/bin/flatpak"):
+            return
+
+        for subcmd in ["remotes", "list"]:
+            r = subprocess.run(
+                    ["/usr/bin/flatpak", subcmd],
+                    stdout=subprocess.PIPE
+            )
+            if r.stdout.decode("utf-8").strip():
+                prevent_auto_removal = True
+                break
+
+        logging.debug("flatpak will{}be marked as manually installed"
+                      .format(" " if prevent_auto_removal else " NOT "))
+
+        if not prevent_auto_removal:
+            return
+
+        self.controller.cache["flatpak"].mark_auto(auto=False)
+
+        for pkg in ("plasma-discover-backend-flatpak",
+                    "gnome-software-plugin-flatpak"):
+            if pkg not in self.controller.cache:
+                continue
+
+            if not self.controller.cache[pkg].is_installed:
+                continue
+
+            logging.debug("{} will be marked as manually installed"
+                          .format(pkg))
+            self.controller.cache[pkg].mark_auto(auto=False)
+
+        self.controller.cache.commit(
+            self._view.getAcquireProgress(),
+            self._view.getInstallProgress(self.controller.cache)
+        )
+
+    def _add_apport_ignore_tracker_extract(self):
+        path = '/etc/apport/blacklist.d/upgrade-quirk-tracker-extract-3'
+
+        try:
+            os.makedirs(os.path.dirname(path), exist_ok=True)
+            with open(path, 'w') as f:
+                f.write('/usr/libexec/tracker-extract-3\n')
+        except Exception as e:
+            logging.debug('Failed to create {}: {}'.format(path, e))
+
+    def _remove_apport_ignore_tracker_extract(self):
+        path = '/etc/apport/blacklist.d/upgrade-quirk-tracker-extract-3'
+
+        try:
+            os.remove(path)
+        except Exception as e:
+            logging.debug('Failed to remove {}: {}'.format(path, e))
+
     def _checkArmCPU(self):
         """
         parse /proc/cpuinfo and search for ARMv6 or greater
@@ -667,7 +742,9 @@ class DistUpgradeQuirks(object):
         # _calculateSnapSizeRequirements call.
         for snap, snap_object in self._snap_list.items():
             command = snap_object['command']
-            if command == 'refresh':
+            if command == 'switch':
+                # TODO: This status should be updated, but the time of
+                # this change to snap switch is post-translation freeze.
                 self._view.updateStatus(_("refreshing snap %s" % snap))
                 popenargs = ["snap", command,
                              "--channel", snap_object['channel'], snap]
@@ -1124,13 +1201,13 @@ class DistUpgradeQuirks(object):
             self._view.processEvents()
             if re.search("^installed: ", snap_info[0], re.MULTILINE):
                 logging.debug("Snap %s is installed" % snap)
-                # its not tracking the release channel so don't refresh
+                # its not tracking the release channel so don't switch
                 if not re.search(r"^tracking:.*%s" % from_channel,
                                  snap_info[0], re.MULTILINE):
                     logging.debug("Snap %s is not tracking the release channel"
                                   % snap)
                     continue
-                snap_object['command'] = 'refresh'
+                snap_object['command'] = 'switch'
             else:
                 # Do not replace packages not installed
                 cache = self.controller.cache
@@ -1187,7 +1264,7 @@ class DistUpgradeQuirks(object):
                             logging.debug("Snap %s is being used by %s. "
                                           "Switching it to stable track"
                                           % (snap, plug_snap))
-                            snap_object['command'] = 'refresh'
+                            snap_object['command'] = 'switch'
                             snap_object['channel'] = 'stable'
                             break
 
@@ -1377,3 +1454,193 @@ class DistUpgradeQuirks(object):
         except IOError as exc:
             logging.error("unable to write new boot config to %s: %s; %s",
                           boot_config_filename, exc, failure_action)
+
+    def _set_generic_font(self):
+        """ Due to changes to the Ubuntu font we enable a generic font
+            (in practice DejaVu or Noto) during the upgrade.
+            See https://launchpad.net/bugs/2034986
+        """
+        temp_font = 'Sans'
+
+        if self._did_change_font:
+            return
+
+        if os.getenv('XDG_SESSION_TYPE', '') in ('', 'tty'):
+            # Avoid running this on server systems or when the upgrade
+            # is done over ssh.
+            return
+
+        if not {'SUDO_UID', 'PKEXEC_UID'}.intersection(os.environ):
+            logging.debug(
+                'Cannot determine how root privileges were gained, will '
+                'not change font'
+            )
+            return
+        try:
+            uid = int(os.getenv('SUDO_UID', os.getenv('PKEXEC_UID')))
+            pwuid = pwd.getpwuid(uid)
+        except ValueError:
+            logging.debug(
+                'Cannot determine non-root UID, will not change font'
+            )
+            return
+
+        schema = 'org.gnome.desktop.interface'
+
+        desktops = os.getenv('XDG_CURRENT_DESKTOP', '').split(':')
+
+        # Some flavors use other schemas for the desktop font.
+        if 'MATE' in desktops or 'UKUI' in desktops:
+            schema = 'org.mate.interface'
+        elif 'X-Cinnamon' in desktops:
+            schema = 'org.cinnamon.desktop.interface'
+
+        # Some flavors lack the systemd integration needed for a
+        # user service, so we create an autostart file instead.
+        use_autostart = bool(
+            set(['Budgie', 'LXQt', 'MATE', 'UKUI', 'X-Cinnamon', 'XFCE'])
+            & set(desktops)
+        )
+
+        r = subprocess.run(
+            ['systemd-run', '--user', '-M', f'{pwuid.pw_name}@.host',
+             '--wait', '--pipe', '-q', '--',
+             '/usr/bin/gsettings', 'get',
+             f'{schema}', 'font-name'],
+            stdout=subprocess.PIPE,
+            encoding='utf-8',
+        )
+
+        (font, _, size) = r.stdout.strip('\'\n').rpartition(' ')
+        font = font or 'Ubuntu'
+        try:
+            int(size)
+        except ValueError:
+            size = '11'
+
+        logging.debug(f'Setting generic font {temp_font} {size} during the '
+                      f'upgrade. Original font is {font} {size}.')
+
+        r = subprocess.run([
+            'systemd-run', '--user', '-M', f'{pwuid.pw_name}@.host',
+            '--wait', '--pipe', '-q', '--',
+            '/usr/bin/gsettings', 'set', f'{schema}',
+            'font-name', f'"{temp_font} {size}"'
+        ])
+        if r.returncode != 0:
+            logging.debug(f'Failed to change font to {temp_font} {size}')
+            return
+
+        self._did_change_font = True
+
+        # Touch a file to indiate that the font should be restored on the next
+        # boot.
+        need_font_restore_file = os.path.join(
+            pwuid.pw_dir, '.config/upgrade-need-font-restore'
+        )
+        os.makedirs(os.path.dirname(need_font_restore_file), exist_ok=True)
+        pathlib.Path(need_font_restore_file).touch(mode=0o666)
+        os.chown(need_font_restore_file, pwuid.pw_uid, pwuid.pw_gid)
+
+        if use_autostart:
+            autostart_file = '/etc/xdg/autostart/upgrade-restore-font.desktop'
+            os.makedirs(os.path.dirname(autostart_file), exist_ok=True)
+            flag = '$HOME/.config/upgrade-need-font-restore'
+            with open(autostart_file, 'w') as f:
+                f.write(
+                    '[Desktop Entry]\n'
+                    'Name=Restore font after upgrade\n'
+                    'Comment=Auto-generated by ubuntu-release-upgrader\n'
+                    'Type=Application\n'
+                    f'Exec=sh -c \'if [ -e "{flag}" ]; then gsettings set '
+                    f'{schema} font-name "{font} {size}";'
+                    f'rm -f "{flag}"; fi\'\n'
+                    'NoDisplay=true\n'
+                )
+            return
+
+        # If we set the font back to normal before a reboot, the font will
+        # still get all messed up. To allow normal usage whether the user
+        # reboots immediately or not, create a service that will run only if a
+        # ~/.config/upgrade-need-font-restore exists, and then remove that file
+        # in ExecStart. This has the effect of creating a one-time service on
+        # the next boot.
+        unit_file = '/usr/lib/systemd/user/upgrade-restore-font.service'
+        os.makedirs(os.path.dirname(unit_file), exist_ok=True)
+
+        with open(unit_file, 'w') as f:
+            f.write(
+                '# Auto-generated by ubuntu-release-upgrader\n'
+                '[Unit]\n'
+                'Description=Restore font after upgrade\n'
+                'After=graphical-session.target dconf.service\n'
+                'ConditionPathExists=%h/.config/upgrade-need-font-restore\n'
+                '\n'
+                '[Service]\n'
+                'Type=oneshot\n'
+                'ExecStart=/usr/bin/gsettings set '
+                f'{schema} font-name \'{font} {size}\'\n'
+                'ExecStart=/usr/bin/rm -f '
+                '%h/.config/upgrade-need-font-restore\n'
+                '\n'
+                '[Install]\n'
+                'WantedBy=graphical-session.target\n'
+            )
+
+        subprocess.run([
+            'systemctl',
+            '--user', '-M', f'{pwuid.pw_name}@.host',
+            'daemon-reload'
+        ])
+        r = subprocess.run([
+            'systemctl',
+            '--user', '-M', f'{pwuid.pw_name}@.host',
+            'enable', os.path.basename(unit_file)
+        ])
+        if r.returncode != 0:
+            logging.debug(f'Failed to enable {os.path.basename(unit_file)}. '
+                          'Font will not be restored on reboot')
+
+    def _test_and_fail_on_bios_with_xfs_boot(self):
+        """
+        If this system is BIOS and has XFS on /boot, do not allow the upgrade
+        to happen (LP: #2039172).
+        """
+        if os.path.isdir('/sys/firmware/efi'):
+            # UEFI boot, nothing to do.
+            return
+
+        boot_fs = None
+        root_fs = None
+        with open('/proc/mounts') as f:
+            for line in f.readlines():
+                try:
+                    (what, where, fs, opts, a, b) = line.strip().split()
+                    where = where.lower()
+                    fs = fs.lower()
+                except ValueError as e:
+                    logging.debug(
+                        f'line \'{line}\' in /proc/mounts not understood'
+                    )
+                    continue
+
+                if where == '/':
+                    root_fs = fs
+                elif where == '/boot':
+                    boot_fs = fs
+
+                if boot_fs and root_fs:
+                    break
+
+        if boot_fs == 'xfs' or (boot_fs is None and root_fs == 'xfs'):
+            # If we got to this point, then this is a BIOS system, and /boot is
+            # XFS or / is XFS and there is no separate /boot partition. Do not
+            # allow the upgrade to happen.
+            logging.error('Cannot upgrade system with BIOS and XFS /boot')
+            self._view.error(
+                _("Cannot upgrade this system"),
+                _("Due to a bug in grub, this system cannot "
+                  "be safely upgraded at this time.\n\n"
+                  "See https://launchpad.net/bugs/2039172.")
+            )
+            self.controller.abort()
